package be.iminds.ilabt.jfed.experimenter_gui.editor.properties.map;

import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Resource;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.ResourceClass;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import de.saring.leafletmap.*;
import javafx.application.Platform;
import javafx.beans.property.ObjectProperty;
import javafx.beans.property.SimpleObjectProperty;
import javafx.concurrent.Worker;
import javafx.event.Event;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.layout.BorderPane;
import javafx.stage.Stage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.util.*;
import java.util.function.BiConsumer;
import java.util.function.Predicate;

/**
 * User: twalcari
 * Date: 1/8/2015
 * Time: 14:34
 */
public class TestbedsMapDialog {
    private static final Logger LOG = LoggerFactory.getLogger(TestbedsMapDialog.class);

    private static final String PROP_GOOGLE_APIKEY = "google.apiKey";
    private static final String API_KEYS_PROPERTIES = "/api-keys.properties";

    private LeafletMapView mapView;

    @FXML
    private BorderPane root;
    @FXML
    private Label testbedNameLabel;
    @FXML
    private Label testbedCountryLabel;
    @FXML
    private Button selectTestbedButton;

    private boolean mapViewIsInitialized = false;

    private ResourceClass resourceClass = null;
    private JFedGuiConfig guiConfig = null;

    private final ObjectProperty<Resource> selectedResource = new SimpleObjectProperty<>();
    private Resource result = null;

    @FXML
    private void initialize() {
        mapView = new LeafletMapView();

        initializeMap();

        root.setCenter(mapView);

        selectTestbedButton.disableProperty().bind(selectedResource.isNull());
    }

    private void initializeMap() {

        MapConfig mapConfig = new MapConfig(
                Collections.singletonList(MapLayer.MAPBOX),
                new ZoomControlConfig(true, ControlPosition.BOTTOM_LEFT),
                new ScaleControlConfig(true, ControlPosition.BOTTOM_RIGHT, true)
        );


        mapView.displayMap(mapConfig)
                .whenCompleteAsync((state, throwable) -> {

                    if (state == Worker.State.SUCCEEDED) {
                        mapViewIsInitialized = true;
                        tryInitializeMap();
                    } else {
                        LOG.error("Map did not initialize successfully: {}", state, throwable);
                    }
                }, Platform::runLater);
    }

    private void tryInitializeMap() {
        if (resourceClass == null) return;
        if (!mapViewIsInitialized) return;


        Predicate<Server> includeServerPredicate = (s) ->
                s.getTestbed() == null || !s.getTestbed().getCategoryNames().contains("offline");
//        includeServerPredicate = (s) -> {
//            assert s.getTestbed() != null;
//            if (s.getTestbed().getCategoryNames().isEmpty())
//                LOG.debug("tryInitializeMap: Testbed \""+s.getTestbed().getId()+"\" has no categories");
//            if (s.getTestbed().getCategoryNames().contains("offline"))
//                LOG.debug("tryInitializeMap: Testbed \""+s.getTestbed().getId()+"\" is offline");
//            return s.getTestbed() == null || !s.getTestbed().getCategoryNames().contains("offline");
//        };

        int serverCount = 0;

        List<Marker> markersList = new ArrayList<>();

        Set<Integer> serverIds = new TreeSet<>();
        for (final Resource curResource : resourceClass.getResources()) {
            Server server = curResource.getServer();
            if (server.getLocation() != null && includeServerPredicate.test(server)) {
                serverIds.add(server.getId());
                serverCount++;

                LatLong authLatLong = new LatLong(server.getLocation().getLatitude(), server.getLocation().getLongitude());
                Marker marker = new Marker(authLatLong, server.getName(), ColorMarkerIcon.BLUE_MARKER, 0);
                marker.setOnClickEventHandler(event -> selectTestbed(curResource));

                markersList.add(marker);

                LOG.debug("Adding marker {} at location {}", server.getName(), server.getLocation());
            }
        }


        //A hack to always show all available locations for "Generic Node"
        if (resourceClass.getId().startsWith("generic-node-")) {
//            {
//                //First add a fake entry for GTS in Copenhagen
//                serverCount++;
//                LatLong authLatLong = new LatLong(55.676098, 12.568337);
//                bounds.extend(authLatLong);
//                MarkerOptions authMarkerOptions = new MarkerOptions();
//                authMarkerOptions.position(authLatLong);
//                authMarkerOptions.title("GTS");
//                Marker authMarker = new Marker(authMarkerOptions);
//                LOG.debug("Adding extra marker for GTS at location lat={} long={}",
//                        authLatLong.getLatitude(), authLatLong.getLongitude());
//                map.addClusterableMarker(authMarker);
//            }

            //then add any server that was not yet added:
            for (Server server : guiConfig.getAllServersInVisibleResources()) {
                if (!serverIds.contains(server.getId()) &&
                        server.getLocation() != null &&
                        includeServerPredicate.test(server)) {
                    serverIds.add(server.getId());
                    serverCount++;

                    LatLong authLatLong =
                            new LatLong(server.getLocation().getLatitude(), server.getLocation().getLongitude());

                    Marker authMarker = new Marker(authLatLong, server.getName(), ColorMarkerIcon.BLUE_MARKER, 0);

                    guiConfig.findAllPossibleResources(server)
                            .stream().findFirst().ifPresent(
                            curResource -> authMarker.setOnClickEventHandler(event -> selectTestbed(curResource))
                    );

                    LOG.debug("Adding extra marker for {} at location lat={} long={}", server.getName(), server.getLocation().getLatitude(), server.getLocation().getLongitude());

                    markersList.add(authMarker);


                }
            }
        }


        MarkerClusterGroup mcg = new MarkerClusterGroup(markersList);
        mapView.addMarkerClusterGroup(mcg);


        LOG.info("Displaying {} servers on the map", serverCount);

        mapView.fitBoundsOfMarkerClusterGroup(mcg);
    }

    private void selectTestbed(Resource resource) {
        Server auth = resource.getServer();
        mapView.panTo(new LatLong(auth.getLocation().getLatitude(), auth.getLocation().getLongitude()));

        testbedNameLabel.setText(auth.getName());
        String address = auth.getLocation().getAddress() != null ? "  Address: " + auth.getLocation().getAddress() : "";
        testbedCountryLabel.setText("Country: " + auth.getLocation().getCountryCode() + address);

        selectedResource.setValue(resource);
    }

    public void setGuiConfig(JFedGuiConfig guiConfig) {
        this.guiConfig = guiConfig;
    }

    public void setCmiManager(ResourceClass resourceClass) {
        this.resourceClass = resourceClass;
        try {
            tryInitializeMap();
        } catch (Throwable t) {
            LOG.error("Failed to initialize map", t);
            throw t;
        }
    }

    @FXML
    private void selectTestbed() {
        result = selectedResource.get();
        ((Stage) selectTestbedButton.getScene().getWindow()).close();
    }

    public Resource getResult() {
        return result;
    }
}
